﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Threading.Tasks;
using System.Net;

namespace TinyMultipartDownloader
{
    class Program
    {
        static void Main(string[] args)
        {
            Task.Run(() => new Downloader().Download(
                "downloade url",
                "client file path"
                )).Wait();

            Console.Read();
        }
    }
    public class Downloader
    {
        public async Task Download(string url, string saveAs)
        {
            using (var httpClient = new HttpClient())
            using (var request = new HttpRequestMessage(HttpMethod.Get, url))
            {
                request.Headers.Range = new RangeHeaderValue(0, 255); // bugfix for 2147483647 exception [ max buffer size => int.MaxValue ]
                // why not HttpCompletionOption.ResponseHeadersRead ?
                // because it will respond 200 only, you won't get partial/206 response
                using (var response = await httpClient.SendAsync(request))
                {
                    File.Create(saveAs).Dispose();
                    if (response.StatusCode == HttpStatusCode.PartialContent)
                    {
                        var contentLength = response.Content.Headers.ContentRange?.Length ?? 0;
                        const double numberOfParts = 5.0;
                        var tasks = new List<Task>();
                        var partSize = (long)Math.Ceiling(contentLength / numberOfParts);
                        for (var i = 0; i < numberOfParts; i++)
                        {
                            var start = i * partSize + Math.Min(1, i);
                            var end = Math.Min((i + 1) * partSize, contentLength);

                            tasks.Add(Task.Run(() => DownloadPart(url, saveAs, start, end)));
                        }

                        await Task.WhenAll(tasks);
                    }
                    else if (response.StatusCode == HttpStatusCode.OK)
                    {
                        await Task.Run(() => DownloadPart(url, saveAs));
                    }
                }
            }
        }

        private async void DownloadPart(string url, string saveAs, long? start = null, long? end = null)
        {
            using (var httpClient = new HttpClient())
            using (var fileStream = new FileStream(saveAs, FileMode.Open, FileAccess.Write, FileShare.Write))
            {
                var message = new HttpRequestMessage(HttpMethod.Get, url);
                if (start.HasValue)
                {
                    message.Headers.Range = new RangeHeaderValue(start, end);
                    fileStream.Position = start.Value;
                }

                Console.WriteLine($"{start}-{end} Started");
                await httpClient.SendAsync(message, HttpCompletionOption.ResponseHeadersRead).Result.Content.CopyToAsync(fileStream);
                Console.WriteLine($"{start}-{end} Downloaded");
            }
        }
    }
}